@@ -60,9 +60,9 @@
 #include "femknot.h"
 #include "feminfo.h"
 #include "femelmer.h"
-#include "femfile.h"
+#include "femfilein.h"
+#include "femfileout.h"
 #include "femfact.h"
-#include "easymesh.h"
 
 
 #define MAXMETHODS 17
@@ -83,8 +83,9 @@ char *IOmethods[] = {
   /*13*/ "GID",
   /*14*/ "GMSH",
   /*15*/ "PARTITIONED",
-  /*16*/ "NASTRAN",
-  /*17*/ "FASTCAP"
+  /*16*/ "UNV",
+  /*17*/ "NASTRAN",
+  /*18*/ "FASTCAP"
 };
 
 
@@ -113,7 +114,8 @@ static void Instructions()
   printf("5)  .inp      : Abaqus input format by Ideas\n");
   printf("6)  .fil      : Abaqus output format\n");
   printf("7)  .FDNEUT   : Gambit (Fidap) neutral file\n");
-  printf("8)  .d        : Easymesh input format\n");
+  printf("8)  .unv      : Universal mesh file format\n");
+  if(0) printf("8)  .d        : Easymesh input format\n");
   printf("9)  .mphtxt   : Comsol Multiphysics mesh format\n");
   printf("10) .dat      : Fieldview format\n");
   printf("11) .node,.ele: Triangle 2D mesh format\n");
@@ -122,7 +124,8 @@ static void Instructions()
   printf("14) .msh      : Gmsh mesh format\n");
   printf("15) .ep.i     : Partitioned ElmerPost format\n");
 #if 0
-  printf("16) .msh      : Nastran format\n");
+  printf("16) .unv      : Universal mesh file format\n");
+  printf("17) .msh      : Nastran format\n");
 #endif 
 
   printf("\nThe second parameter defines the output file format:\n");
@@ -132,8 +135,8 @@ static void Instructions()
 #if 0
   printf("5)  .inp      : Abaqus input format\n");
   printf("7)  .fidap    : Fidap format\n");
-  printf("8)  .n .e .s  : Easymesh output format\n");
-  printf("17) .ep       : Fastcap input format.\n");
+  if(0) printf("8)  .n .e .s  : Easymesh output format\n");
+  printf("18) .ep       : Fastcap input format.\n");
 #endif
 
   printf("\nThe third parameter is the name of the input file.\n");
@@ -149,6 +152,7 @@ static void Instructions()
   printf("-triangles           : rectangles will be divided to triangles\n");
   printf("-merge real          : merges nodes that are close to each other\n");
   printf("-order real[3]       : reorder elements and nodes using c1*x+c2*y+c3*z\n");
+  printf("-centralize          : set the center of the mesh to origin\n");
   printf("-scale real[3]       : scale the coordinates with vector real[3]\n");
   printf("-translate real[3]   : translate the nodes with vector real[3]\n");
   printf("-rotate real[3]      : rotate around the main axis with angles real[3]\n");
@@ -159,22 +163,12 @@ static void Instructions()
   printf("-polar real          : map 2D mesh to a cylindrical shell with given radius\n");
   printf("-cylinder            : map 2D/3D cylindrical mesh to a cartesian mesh\n");
   printf("-reduce int[2]       : reduce element order at material interval [int1 int2]\n");
-  printf("-increase            : increase element order from one to two\n");
-#if 0
-  /* This functionality has moved into the ElmerSolver */
-  printf("-pelem int[3]        : p-elements of power int3 at interval [int1 int2]\n");
-  printf("-belem int[3]        : set bubble dofs to int3 at interval [int1 int2]\n");
-#endif
-  printf("-partition int[3]    : the mesh will be partitioned in main directions\n");
-  printf("-partorder real[3]   : in the above method, the direction of the ordering\n");
-#if PARTMETIS
-  printf("-metis int[2]        : the mesh will be partitioned with Metis\n");
-#endif
-  printf("-periodic int[3]     : decleare the periodic coordinate directions\n");
+  printf("-increase            : increase element order from linear to quadratic\n");
   printf("-bcoffset int        : add an offset to the boundary conditions\n");
   printf("-discont int         : make the boundary to have secondary nodes\n");
   printf("-connect int         : make the boundary to have internal connection among its elements\n");
   printf("-removelowdim        : remove boundaries that are two ranks lower than highest dim\n");
+  printf("-removeunused        : remove nodes that are not used in any element\n");
   printf("-bulkorder           : renumber materials types from 1 so that every number is used\n");
   printf("-boundorder          : renumber boundary types from 1 so that every number is used\n");
   printf("-autoclean           : this performs the united action of the three above\n");
@@ -188,10 +182,28 @@ static void Instructions()
   printf("-3d / -2d / -1d      : mesh is 3, 2 or 1-dimensional (applies to examples)\n");
   printf("-isoparam            : ensure that higher order elements are convex\n");
   printf("-nobound             : disable saving of boundary elements in ElmerPost format\n");
-  printf("-names               : conserve name information where applicable\n");
+
+  printf("\nThe following keywords are related only to the parallel Elmer computations.\n");
+  printf("-partition int[4]    : the mesh will be partitioned in main directions\n");
+  printf("-partorder real[3]   : in the above method, the direction of the ordering\n");
+#if PARTMETIS
+  printf("-metis int[2]        : the mesh will be partitioned with Metis\n");
+#endif
+  printf("-halo                : create halo for the partitioning\n");
+  printf("-indirect            : create indirect connections in the partitioning\n");
+  printf("-periodic int[3]     : decleare the periodic coordinate directions for parallel meshes\n");
+  printf("-saveinterval int[3] : the first, last and step for fusing parallel data\n");
+
+  if(0) printf("-names               : conserve name information where applicable\n");
 #if 0
   printf("-map str             : file with mapping info for mesh-to-mesh interpolation\n");
 #endif
+#if 0
+  /* This functionality has moved into the ElmerSolver */
+  printf("-pelem int[3]        : p-elements of power int3 at interval [int1 int2]\n");
+  printf("-belem int[3]        : set bubble dofs to int3 at interval [int1 int2]\n");
+#endif
+
 }
 
 
@@ -223,15 +235,20 @@ void InitParameters(struct ElmergridType *eg)
   eg->elements3d = 0;
   eg->nodes3d = 0;
   eg->metis = 0;
+  eg->partitionhalo = FALSE;
+  eg->partitionindirect = FALSE;
   eg->reduce = FALSE;
   eg->increase = FALSE;
   eg->translate = FALSE;
   eg->isoparam = FALSE;
   eg->removelowdim = FALSE;
+  eg->removeunused = FALSE;
   eg->dim = 3;
+  eg->center = FALSE;
   eg->scale = FALSE;
   eg->order = FALSE;
   eg->boundbounds = 0;
+  eg->saveinterval[0] = eg->saveinterval[1] = eg->saveinterval[2] = 0;
   eg->bulkbounds = 0;
   eg->partorder = FALSE;
   eg->findsides = FALSE;
@@ -338,7 +355,7 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
     if(strcmp(argv[arg],"-merge") == 0) {
       if(arg+1 >= argc) {
 	printf("Give a parameter for critical distance.\n");
-	return(18);
+	return(3);
       }
       else {
 	eg->merge = TRUE;
@@ -349,7 +366,7 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
     if(strcmp(argv[arg],"-order") == 0) {
       if(arg+dim >= argc) {
 	printf("Give %d parameters for the order vector.\n",dim);
- 	return(11);
+ 	return(4);
       }
       else {
 	eg->order = TRUE;
@@ -363,10 +380,22 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
       eg->order = 2;
     }
 
+    if(strcmp(argv[arg],"-halo") == 0) {
+      eg->partitionhalo = TRUE;
+    }
+    if(strcmp(argv[arg],"-indirect") == 0) {
+      eg->partitionindirect = TRUE;
+    }
+    if(strcmp(argv[arg],"-metisorder") == 0) {
+      eg->order = 3;
+    }
+    if(strcmp(argv[arg],"-centralize") == 0) {
+      eg->center = TRUE;
+    }
     if(strcmp(argv[arg],"-scale") == 0) {
       if(arg+dim >= argc) {
 	printf("Give %d parameters for the scaling.\n",dim);
- 	return(12);
+ 	return(5);
      }
       else {
 	eg->scale = TRUE;
@@ -379,7 +408,7 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
     if(strcmp(argv[arg],"-translate") == 0) {
       if(arg+dim >= argc) {
 	printf("Give %d parameters for the translate vector.\n",dim);
-	return(17);
+	return(6);
       }
       else {
 	eg->translate = TRUE;
@@ -389,10 +418,22 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
       }
     }
 
+    if(strcmp(argv[arg],"-saveinterval") == 0) {
+      if(arg+dim >= argc) {
+	printf("Give min, max and step for the interval.\n");
+	return(7);
+      }
+      else {
+	eg->saveinterval[0] = atoi(argv[arg+1]);
+	eg->saveinterval[1] = atoi(argv[arg+2]);
+	eg->saveinterval[2] = atoi(argv[arg+3]);
+      }
+    }
+
     if(strcmp(argv[arg],"-rotate") == 0 || strcmp(argv[arg],"-rotate") == 0) {
       if(arg+dim >= argc) {
 	printf("Give three parameters for the rotation angles.\n");
-	return(16);
+	return(8);
       }
       else {
 	eg->rotate = TRUE;
@@ -405,7 +446,7 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
     if(strcmp(argv[arg],"-clone") == 0) {
       if(arg+dim >= argc) {
 	printf("Give the number of clones in each %d directions.\n",dim);
- 	return(13);
+ 	return(9);
      }
       else {
 	eg->clone[0] = atoi(argv[arg+1]);
@@ -416,7 +457,7 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
     if(strcmp(argv[arg],"-clonesize") == 0) {
       if(arg+dim >= argc) {
 	printf("Give the clone size in each %d directions.\n",dim);
- 	return(15);
+ 	return(10);
       }
       else {
 	eg->clonesize[0] = atof(argv[arg+1]);
@@ -438,7 +479,7 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
     if(strcmp(argv[arg],"-mirrorbc") == 0) {
       if(arg+1 >= argc) {
 	printf("Give the number of symmetry BC.\n");
- 	return(14);
+ 	return(11);
       }
       else {
 	eg->mirrorbc = atoi(argv[arg+1]);
@@ -460,12 +501,19 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
       printf("Lower dimensional boundaries will be removed\n");
     }   
 
+    if(strcmp(argv[arg],"-removeunused") == 0) {
+      eg->removeunused = TRUE;
+      printf("Nodes that do not appear in any element will be removed\n");
+    }   
+
     if(strcmp(argv[arg],"-autoclean") == 0) {
       eg->removelowdim = TRUE;
       eg->bulkorder = TRUE;
       eg->boundorder = TRUE;
+      eg->removeunused = TRUE;
       printf("Lower dimensional boundaries will be removed\n");
       printf("Materials and boundaries will be renumbered\n");
+      printf("Nodes that do not appear in any element will be removed\n");
     }   
 
     if(strcmp(argv[arg],"-polar") == 0) {
@@ -488,7 +536,7 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
     if(strcmp(argv[arg],"-reduce") == 0) {
       if(arg+2 >= argc) {
 	printf("Give two material for the interval.\n");
- 	return(8);
+ 	return(12);
       }
       else {
 	eg->reduce = TRUE;      
@@ -518,7 +566,7 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
     if(strcmp(argv[arg],"-partition") == 0) {
       if(arg+dim >= argc) {
 	printf("The number of partitions in %d dims is required as paramaters.\n",dim);
-	return(3);
+	return(13);
       }
       else {
 	eg->partitions = 1;
@@ -530,7 +578,10 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
 	  if(eg->partdim[i] == 0) eg->partdim[i] = 1;
 	  eg->partitions *= eg->partdim[i];
 	}
-	if(eg->partdim[0] < 0) eg->partdim[0] = -eg->partdim[0];
+	eg->partopt = 0;
+	if(arg+4 < argc) 
+	  if(argv[arg+4][0] != '-') eg->partopt = atoi(argv[arg+4]);
+
 	printf("The mesh will be partitioned with simple division to %d partitions.\n",
 	       eg->partitions);
       }
@@ -538,7 +589,7 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
     if(strcmp(argv[arg],"-partorder") == 0) {
       if(arg+dim >= argc) {
 	printf("Give %d parameters for the order vector.\n",dim);
- 	return(4);
+ 	return(14);
       }
       else {
 	eg->partorder = 1;
@@ -552,17 +603,14 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
 #if PARTMETIS
       if(arg+1 >= argc) {
 	printf("The number of partitions is required as a parameter\n");
-	return(5);
+	return(15);
       }
       else {
 	eg->metis = atoi(argv[arg+1]);
 	printf("The mesh will be partitioned with Metis to %d partitions.\n",eg->metis);
-	if(arg+2 < argc) { 
-	  if(argv[arg+2][0] != '-') eg->metisopt = atoi(argv[arg+2]);
-	}
-	else {
-	  eg->metisopt = 0;
-	}
+	eg->partopt = 0;
+	if(arg+2 < argc) 
+	  if(argv[arg+2][0] != '-') eg->partopt = atoi(argv[arg+2]);
       }
 #else
       printf("This version of ElmerGrid was compiled without Metis library!\n");
@@ -572,7 +620,7 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
     if(strcmp(argv[arg],"-periodic") == 0) {
       if(arg+dim >= argc) {
 	printf("Give the periodic coordinate directions (e.g. 1 1 0)\n");
- 	return(9);
+ 	return(16);
       }
       else {
 	eg->periodicdim[0] = atoi(argv[arg+1]);
@@ -584,7 +632,7 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
     if(strcmp(argv[arg],"-discont") == 0) {
       if(arg+1 >= argc) {
 	printf("Give the discontinuous boundary conditions.\n");
- 	return(10);
+ 	return(17);
       }
       else {
 	eg->discontbounds[eg->discont] = atoi(argv[arg+1]);
@@ -629,11 +677,11 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
     if(strcmp(argv[arg],"-layer") == 0) {
       if(arg+4 >= argc) {
 	printf("Give four parameters for the layer: boundary, elements, thickness, ratio.\n");
-	return(19);
+	return(18);
       }
       else if(eg->layers == MAXBOUNDARIES) {
 	printf("There can only be %d layers, sorry.\n",MAXBOUNDARIES);
-	return(20);
+	return(19);
       }
       else {
 	eg->layerbounds[eg->layers] = atoi(argv[arg+1]);
@@ -648,7 +696,7 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
     if(strcmp(argv[arg],"-layermove") == 0) {
       if(arg+1 >= argc) {
 	printf("Give maximum number of Jacobi filters.\n");
- 	return(11);
+ 	return(20);
       }
       else {
 	eg->layermove = atoi(argv[arg+1]);
@@ -660,11 +708,11 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
     if(strcmp(argv[arg],"-divlayer") == 0) {
       if(arg+4 >= argc) {
 	printf("Give four parameters for the layer: boundary, elements, relative thickness, ratio.\n");
-	return(19);
+	return(21);
       }
       else if(abs(eg->layers) == MAXBOUNDARIES) {
 	printf("There can only be %d layers, sorry.\n",MAXBOUNDARIES);
-	return(20);
+	return(22);
       }
       else {
 	eg->layerbounds[abs(eg->layers)] = atoi(argv[arg+1]);
@@ -698,7 +746,7 @@ int InlineParameters(struct ElmergridType *eg,int argc,char *argv[])
     if(strcmp(argv[arg],"-map") ==0) {
       if(arg+1 >= argc) {
 	printf("Give the name of the mapping file\n");
-	return(7);
+	return(23);
       }
       else {
 	strcpy(eg->mapfile,argv[arg+1]);
@@ -757,25 +805,24 @@ int LoadCommands(char *prefix,struct ElmergridType *eg,
   int i,j,k,l,error=0;
 
   if( mode == 0) {  
-    if ((in = fopen("ELMERGRID_STARTINFO","r")) == NULL) {
-      return(1);
-    }
-    else {
+    if (in = fopen("ELMERGRID_STARTINFO","r")) {
       fscanf(in,"%s",filename);
       fclose(in);
       printf("Using the file %s defined in ELMERGRID_STARTINFO\n",filename);
       if ((in = fopen(filename,"r")) == NULL) {
 	printf("LoadCommands: opening of the file '%s' wasn't succesfull !\n",filename);
-	return(2);
+	return(1);
       }    
+      else printf("Loading ElmerGrid commands from file '%s'.\n",filename);    
     }    
-    if(info) printf("Loading ElmerGrid commands from file '%s'.\n",filename);    
+    else 
+      return(2);
   }
   if(mode == 1) { 
     AddExtension(prefix,filename,"eg");
     if ((in = fopen(filename,"r")) == NULL) {
       printf("LoadCommands: opening of the file '%s' wasn't succesfull !\n",filename);
-      return(2);
+      return(3);
     }    
     if(info) printf("Loading ElmerGrid commands from file '%s'.\n",filename);    
   }
@@ -783,7 +830,7 @@ int LoadCommands(char *prefix,struct ElmergridType *eg,
     AddExtension(prefix,filename,"grd");
     if ((in = fopen(filename,"r")) == NULL) {
       printf("LoadCommands: opening of the file '%s' wasn't succesfull !\n",filename);
-      return(3);
+      return(4);
     }    
     if(info) printf("Loading ElmerGrid commands from file '%s'.\n",filename);
   }
@@ -799,7 +846,7 @@ int LoadCommands(char *prefix,struct ElmergridType *eg,
 
     /* If the mode is the command file mode read also the file information from the command file. */
 
-    if(mode == 0) {
+    if(mode <= 1) {
       if(strstr(command,"INPUT FILE")) {
 	sscanf(params,"%s",&eg->filesin[0]);
       }
@@ -869,6 +916,9 @@ int LoadCommands(char *prefix,struct ElmergridType *eg,
       else if(eg->dim == 3) 
 	sscanf(params,"%le%le%le",&eg->cscale[0],&eg->cscale[1],&eg->cscale[2]);
     }
+    else if(strstr(command,"CENTRALIZE")) {
+      eg->center = TRUE;
+    }
     else if(strstr(command,"TRANSLATE")) {
       eg->translate = TRUE;
       if(eg->dim == 1) 
@@ -968,7 +1018,7 @@ int LoadCommands(char *prefix,struct ElmergridType *eg,
     }
     else if(strstr(command,"METIS OPTION")) {
 #if PARTMETIS
-      sscanf(params,"%d",&eg->metisopt);
+      sscanf(params,"%d",&eg->partopt);
 #else
       printf("This version of ElmerGrid was compiled without Metis library!\n");
 #endif
@@ -1000,7 +1050,14 @@ int LoadCommands(char *prefix,struct ElmergridType *eg,
       if(eg->dim == 3) sscanf(params,"%d%d%d",&eg->periodicdim[0],
 			      &eg->periodicdim[1],&eg->periodicdim[2]);
     }
-
+    else if(strstr(command,"HALO")) {
+      for(j=0;j<MAXLINESIZE;j++) params[j] = toupper(params[j]);
+      if(strstr(params,"TRUE")) eg->partitionhalo = TRUE;      
+    }
+    else if(strstr(command,"INDIRECT")) {
+      for(j=0;j<MAXLINESIZE;j++) params[j] = toupper(params[j]);
+      if(strstr(params,"TRUE")) eg->partitionindirect = TRUE;      
+    }
     else if(strstr(command,"BOUNDARY BOUNDARY")) {
       for(i=0;i<MAXBOUNDARIES;i++) {
 	if(i>0) Getline(params,in);
@@ -1073,6 +1130,10 @@ int LoadCommands(char *prefix,struct ElmergridType *eg,
       for(j=0;j<MAXLINESIZE;j++) params[j] = toupper(params[j]);
       if(strstr(params,"TRUE")) eg->removelowdim = TRUE; 
     }
+    else if(strstr(command,"REMOVE UNUSED NODES")) {
+      for(j=0;j<MAXLINESIZE;j++) params[j] = toupper(params[j]);
+      if(strstr(params,"TRUE")) eg->removeunused = TRUE; 
+    }
     else if(strstr(command,"REORDER MATERIAL")) {
       for(j=0;j<MAXLINESIZE;j++) params[j] = toupper(params[j]);
       if(strstr(params,"TRUE")) eg->bulkorder = TRUE; 
@@ -1140,7 +1201,7 @@ int main(int argc, char *argv[])
 {
   int i,j,k,l,inmethod,outmethod,info,errorstat,sides;
   int nogrids,nomeshes,nofile,dim,elementsredone=0;
-  int nodes3d,elements3d;
+  int nodes3d,elements3d,showmem;
   Real mergeeps;
   char prefix[MAXFILESIZE];
   struct GridType *grids;
@@ -1150,7 +1211,7 @@ int main(int argc, char *argv[])
   struct ElmergridType eg;
   long ii;
 
-
+  showmem = TRUE;
   printf("\nStarting program Elmergrid\n");
 
   InitParameters(&eg);
@@ -1158,7 +1219,12 @@ int main(int argc, char *argv[])
   InitGrid(grids);
   info = TRUE;
 
-  if(argc <= 2) {
+  if(argc <= 1) {
+    errorstat = LoadCommands(argv[1],&eg,grids,argc-1,info);     
+    Instructions();
+    if(errorstat) Goodbye();
+  }
+  if(argc == 2) {
     errorstat = LoadCommands(argv[1],&eg,grids,argc-1,info);     
     if(errorstat) Goodbye();
   }
@@ -1183,10 +1249,12 @@ int main(int argc, char *argv[])
       Goodbye();
     }
   }
+#if 0
   if(eg.inmethod != 8 && eg.outmethod == 5) {
     printf("To write Easymesh format you need to read easymesh format!\n");
     errorstat++;
   }
+#endif
 
 
   /**********************************/
@@ -1269,13 +1337,14 @@ int main(int argc, char *argv[])
       boundaries[nofile][i].created = FALSE; 
       boundaries[nofile][i].nosides = 0;
     }
-    if(!eg.usenames) data[nofile].boundarynamesexist = data[nofile].bodynamesexist = FALSE;
+    if(0 && !eg.usenames) data[nofile].boundarynamesexist = data[nofile].bodynamesexist = FALSE;
     ElementsToBoundaryConditions(&(data[nofile]),boundaries[nofile],TRUE);
     RenumberBoundaryTypes(&data[nofile],boundaries[nofile],TRUE,0,info);
   
     nomeshes++;
     break;
 
+#if 0
   case 8: 
     InitializeKnots(&(data[nofile]));
     if( Easymesh(argc,argv,&data[nofile].noknots,
@@ -1297,6 +1366,7 @@ int main(int argc, char *argv[])
       Goodbye();    
     nomeshes++;
     break;
+#endif
 
  case 9:
     boundaries[nofile] = (struct BoundaryType*)
@@ -1382,13 +1452,28 @@ int main(int argc, char *argv[])
     break;
 
   case 15: 
-    if(info) printf("Partitioned solution is fused on-the-fly therefore no operation may be performed.\n");
-    FuseSolutionElmerPartitioned(eg.filesin[nofile],eg.filesout[nofile],eg.decimals,info);
+    if(info) printf("Partitioned solution is fused on-the-fly therefore no other operations may be performed.\n");
+    FuseSolutionElmerPartitioned(eg.filesin[nofile],eg.filesout[nofile],eg.decimals,
+				 eg.saveinterval[0],eg.saveinterval[1],eg.saveinterval[2],info);
     if(info) printf("Finishing with the fusion of partitioned Elmer solutions\n");
     Goodbye();
     break;
 
+  case 8:
   case 16:
+    boundaries[nofile] = (struct BoundaryType*)
+      malloc((size_t) (MAXBOUNDARIES)*sizeof(struct BoundaryType)); 	
+    for(i=0;i<MAXBOUNDARIES;i++) {
+      boundaries[nofile][i].created = FALSE; 
+      boundaries[nofile][i].nosides = 0;
+    }
+    if (LoadUniversalMesh(&(data[nofile]),eg.filesin[nofile],TRUE))
+      Goodbye();
+    if(1) ElementsToBoundaryConditions(&(data[nofile]),boundaries[nofile],TRUE);
+    nomeshes++;
+    break;
+
+  case 17:
     boundaries[nofile] = (struct BoundaryType*)
       malloc((size_t) (MAXBOUNDARIES)*sizeof(struct BoundaryType)); 	
     for(i=0;i<MAXBOUNDARIES;i++) {
@@ -1400,7 +1485,6 @@ int main(int argc, char *argv[])
     nomeshes++;
     break;
 
-
   default:
     Instructions();
     Goodbye();
@@ -1419,7 +1503,6 @@ int main(int argc, char *argv[])
 
   printf("\nElmergrid creating and manipulating meshes:\n");
 
-
   if(nogrids > nomeshes && outmethod != 1) { 
 
     nomeshes = nogrids;
@@ -1605,8 +1688,11 @@ int main(int argc, char *argv[])
   for(k=0;k<nomeshes;k++) {
     if(eg.merge) 
       MergeElements(&data[k],boundaries[k],eg.order,eg.corder,eg.cmerge,FALSE,TRUE);
+    else if(eg.order == 3) 
+      ReorderElementsMetis(&data[k],TRUE);
     else if(eg.order) 
       ReorderElements(&data[k],boundaries[k],eg.order,eg.corder,TRUE);
+    
     if(eg.isoparam) 
       IsoparametricElements(&data[k],boundaries[k],TRUE,info);
   }  
@@ -1626,31 +1712,23 @@ int main(int argc, char *argv[])
 	for(l=0;l<eg.bulkbounds;l++) {
 	  FindBulkBoundary(&data[k],eg.bulkbound[3*l],eg.bulkbound[3*l+1],
 			   boundnodes,&noboundnodes,info);
-	  FindNewBoundaries(&data[k],boundaries[k],boundnodes,eg.bulkbound[3*l+2],info);
+	  FindNewBoundaries(&data[k],boundaries[k],boundnodes,eg.bulkbound[3*l+2],1,info);
 	}
       }
       if(eg.boundbounds) {
 	for(l=0;l<eg.boundbounds;l++) {	
 	  FindBoundaryBoundary(&data[k],boundaries[k],eg.boundbound[3*l],eg.boundbound[3*l+1],
 			       boundnodes,&noboundnodes,info);
-	  FindNewBoundaries(&data[k],boundaries[k],boundnodes,eg.boundbound[3*l+2],info);
+	  FindNewBoundaries(&data[k],boundaries[k],boundnodes,eg.boundbound[3*l+2],2,info);
 	}
       }
       free_Ivector(boundnodes,1,data[k].noknots);
     }
 
+#if 0
     if(eg.bulkbounds || eg.boundbounds)
       SeparateCartesianBoundaries(&data[k],boundaries[k],info);
-  }
-
-
-  /* Create a variable so that when saving data in ElmerPost format there is something to visualize */
-  for(k=0;k<nomeshes;k++) {
-    if(data[k].variables == 0) {
-      CreateVariable(&data[k],1,1,0.0,"Number",FALSE);
-      for(i=1;i<=data[k].alldofs[1];i++)
-	data[k].dofs[1][i] = (Real)(i);
-    }
+#endif
   }
 
   for(k=0;k<nomeshes;k++) 
@@ -1660,6 +1738,10 @@ int main(int argc, char *argv[])
     for(k=0;k<nomeshes;k++)
       RemoveLowerDimensionalBoundaries(&data[k],boundaries[k],info);
 
+  if(eg.removeunused) 
+    for(k=0;k<nomeshes;k++)
+      RemoveUnusedNodes(&data[k],info);
+
   if(eg.boundorder || eg.bcoffset) 
     for(k=0;k<nomeshes;k++) 
       RenumberBoundaryTypes(&data[k],boundaries[k],eg.boundorder,eg.bcoffset,info);
@@ -1691,7 +1773,7 @@ int main(int argc, char *argv[])
 	}
       }
     }
-    printf("Renumbering boundary types finished\n");
+    if(info) printf("Renumbering boundary types finished\n");
   }
   
   if(eg.bulkmappings) {
@@ -1710,37 +1792,36 @@ int main(int argc, char *argv[])
 	}
       }
     }
-    printf("Renumbering material indexes finished\n");
+    if(info) printf("Renumbering material indexes finished\n");
   }
 
-
   for(k=0;k<nomeshes;k++) 
     if(eg.periodicdim[0] || eg.periodicdim[1] || eg.periodicdim[2]) 
       FindPeriodicNodes(&data[k],eg.periodicdim,info);
-
    
   for(k=0;k<nomeshes;k++) {
-    if(eg.partitions > 1) 
-      PartitionSimpleElements(&data[k],eg.partdim,eg.periodicdim,eg.partorder,eg.partcorder,info);	
-    else if(eg.partitions < -1) { 
-      eg.partitions = abs( eg.partitions );
-      PartitionSimpleNodes(&data[k],eg.partdim,eg.periodicdim,eg.partorder,eg.partcorder,info);	
+    int noopt = 0;
+    if(eg.partitions) {
+      if(eg.partopt % 2 == 0) 
+	PartitionSimpleElements(&data[k],eg.partdim,eg.periodicdim,eg.partorder,eg.partcorder,info);	
+      else 
+	PartitionSimpleNodes(&data[k],eg.partdim,eg.periodicdim,eg.partorder,eg.partcorder,info);	
+      noopt = eg.partopt / 2;      
     }
 #if PARTMETIS
     if(eg.metis) {
-      if(eg.metisopt % 4 == 0) 
-	PartitionMetisElements(&data[k],eg.metis,info);
+      if(eg.partopt % 5 <= 1) 
+	PartitionMetisElements(&data[k],eg.metis,eg.partopt % 5,info);
       else
-	PartitionMetisNodes(&data[k],eg.metis,eg.metisopt % 4,info);      
+	PartitionMetisNodes(&data[k],eg.metis,eg.partopt % 5,info);      
+      noopt = eg.partopt / 5;      
     }
 #endif
-    if(eg.partitions || eg.metis ) {
-      int noopt = 0;
-      if(eg.metis) noopt = eg.metisopt / 4;
+    if(eg.partitions || eg.metis ) 
       OptimizePartitioning(&data[k],boundaries[k],noopt,info);
-    }
   }
 
+
   if(eg.pelems || eg.belems || eg.advancedmat) {
     int currenttype;
 
@@ -1785,7 +1866,7 @@ int main(int argc, char *argv[])
 	      currenttype = -1;
 	    }
 	}
-	printf("Creating bubble elements finished\n");
+	if(info) printf("Creating bubble elements finished\n");
       }
 
       if(eg.advancedmat) {
@@ -1797,7 +1878,7 @@ int main(int argc, char *argv[])
 
 	  for(i=1;i<=6;i++) {
 	    if(eg.advancedelem[7*l+i] < 0 || eg.advancedelem[7*l+i] > 99) {
-	      printf("Advanced elements limited to 9 or 99 (not %d)!\n",eg.advancedelem[7*l+i]);
+	      if(info) printf("Advanced elements limited to 9 or 99 (not %d)!\n",eg.advancedelem[7*l+i]);
 	    }
 	  }
 
@@ -1809,7 +1890,7 @@ int main(int argc, char *argv[])
 	    }
 	  }
 	}
-	printf("Creating advanced elements finished\n");
+	if(info) printf("Creating advanced elements finished\n");
       }
     }
   }
@@ -1826,22 +1907,43 @@ int main(int argc, char *argv[])
   case 2:
     for(k=0;k<nomeshes;k++) {
       if(data[k].nopartitions > 1) 
-	SaveElmerInputPartitioned(&data[k],boundaries[k],eg.filesout[k],eg.decimals,info);
+	SaveElmerInputPartitioned(&data[k],boundaries[k],eg.filesout[k],eg.decimals,
+				  eg.partitionhalo,eg.partitionindirect,info);
       else
-	SaveElmerInput(&data[k],boundaries[k],eg.filesout[k],eg.decimals,0,info);
+	SaveElmerInput(&data[k],boundaries[k],eg.filesout[k],eg.decimals,info);
     }
     break;
 
+  case 22:
+
+    for(k=0;k<nomeshes;k++) {
+      SaveElmerInputFemBem(&data[k],boundaries[k],eg.filesout[k],eg.decimals,info);
+    }
+    break;
+
+
   case 3:
-    for(k=0;k<nomeshes;k++) 
+      /* Create a variable so that when saving data in ElmerPost format there is something to visualize */
+    for(k=0;k<nomeshes;k++) {
+      if(data[k].variables == 0) {
+	CreateVariable(&data[k],1,1,0.0,"Number",FALSE);
+	for(i=1;i<=data[k].alldofs[1];i++)
+	  data[k].dofs[1][i] = (Real)(i);	
+      }
       SaveSolutionElmer(&data[k],boundaries[k],eg.saveboundaries ? MAXBOUNDARIES:0,
 			eg.filesout[k],eg.decimals,info);
+    }
     break;
 
   case 4:
     printf("The output number 4 still refers to ep-file but will become obsolite in time\n");
     printf("Rather use number 3 for ElmerPost output format\n");
     for(k=0;k<nomeshes;k++) {
+      if(data[k].variables == 0) {
+	CreateVariable(&data[k],1,1,0.0,"Number",FALSE);
+	for(i=1;i<=data[k].alldofs[1];i++)
+	  data[k].dofs[1][i] = (Real)(i);	      
+      }
       SaveSolutionElmer(&data[k],boundaries[k],eg.saveboundaries ? MAXBOUNDARIES:0,
 			eg.filesout[k],eg.decimals,info);
     }
@@ -1857,11 +1959,13 @@ int main(int argc, char *argv[])
       SaveFidapOutput(&data[k],eg.filesout[k],info,1,data[k].dofs[1]);
     break;
 
+#if 0
   case 8:
     EasymeshSave();
     break;
+#endif
 
-  case 17:    
+  case 18:    
     for(k=0;k<nomeshes;k++) 
       SaveFastcapInput(&data[k],boundaries[k],eg.filesout[k],eg.decimals,info);
     break;
